/*========================================================================
 * Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved..
 *
 * Name - shmem.c
 *
 * Description - Verify that shared memory is functional on a Unix host
 *
 * $Id: shmem.c,v 1.10 2008-01-09 19:03:03 stever Exp $
 *
 *========================================================================
 */

#include <flag.ht> /* MUST BE FIRST! */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h> /* for access() */

#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>

#if defined(FLG_NO_SEMUN)
union semun {
  int             val;
  struct semid_ds *buf;
  ushort          *array;
  };
#endif

#if defined(FLG_HPUX_UNIX) 
  /* hpux 9.0 needs a definition, 10.0 has one, but we'll live without it. */
#elif defined(FLG_SUN_UNIX)
  key_t ftok(const char *, char);
#elif defined(FLG_SEQUENT_UNIX)  || defined(FLG_ULTRIX_UNIX)
  key_t ftok(const char *, char);
#else
  /* rely on system prototypes */
#endif

#undef DOIT
#if defined(FLG_SUN_UNIX) && defined(FLG_GNUC)
#define DOIT TRUE
#endif
#if defined(FLG_ULTRIX_UNIX)
#define DOIT TRUE
#endif

#if defined(DOIT)
#undef DOIT
  int semctl(int semid, int semnum, int cmd, union semun arg);
#elif defined(FLG_HPUX_UNIX)
  int semctl(int semid , int semnum , int cmd, ...);
#elif defined(FLG_AIX_UNIX)
  int semctl(int semid , int semnum , int cmd, ...);
#else
  /* rely on system prototypes */
#endif

int semget(key_t key, int nsems, int semflg);

#if defined(RISC6000AIX)
  /* prototypes OK */
#elif defined(FLG_HPUX_UNIX)
  /* prototypes OK */
#elif defined(FLG_APPLE_UNIX)

#else
int semop(int semid, struct sembuf *sops, unsigned int nsops);
#endif

#if defined(FLG_SUN_UNIX)
  void *shmat(int, const void *, int);
#elif defined(FLG_HPUX_UNIX)
  void *shmat(int, const void *, int);
#elif defined(FLG_ULTRIX_UNIX)
  char *shmat(int shmid, char *shmaddr, int shmflg);
#else
  /* rely on system prototypes */
#endif

#if defined(FLG_NCR_UNIX)
  /* use system prototypes */
#elif defined(FLG_SOLARIS_UNIX)
  /* use system prototypes */
#elif defined(FLG_APPLE_UNIX)

#else
int shmctl(int shmid, int cmd, struct shmid_ds *buf);
#endif

#if defined(FLG_SUN_UNIX) /* gcc, actually */ || defined(FLG_ULTRIX_UNIX)
  int shmdt(const void *);
#elif defined(FLG_HPUX_UNIX)
  int shmdt(const void *);
#else
  /* rely on system prototypes */
#endif

#if defined(FLG_HPUX_UNIX)
  /* prototypes ok */
#elif defined(FLG_SOLARIS_UNIX)
  /* prototypes ok */
#elif defined(FLG_AIX_UNIX)
  /* prototypes ok */
#elif defined(FLG_LINUX_UNIX)
  /* prototypes ok */

#elif defined(FLG_APPLE_UNIX)

#else
int shmget(key_t key, int size, int shmflg);
#endif

#if !defined(SHM_R)
#define SHM_R 0400   /* user read permission */
#endif
#if !defined(SHM_W)
#define SHM_W 0200   /* user write permission */
#endif

/* the group permissions are not not defined at all in shm.h */
#define SHMG_R 0040  /* group read permission */
#define SHMG_W 0020  /* group write permission */

/* define permissions used for our system */
#define SHM_PERMISSIONS  (SHM_R | SHM_W | SHMG_R | SHMG_W)

#ifdef FLG_HPUX_UNIX
#define SHMGETCAST(x) (size_t)x
#else
#define SHMGETCAST(x) x
#endif

static void showErrno(int num)
{
#if defined(FLG_SUN_UNIX)
  printf("  errno = %d\n", num);
#else
  printf("  errno = %d (%s)\n", num, strerror(num));
#endif
}

static int testMem(const char *path, int segSize, int numProcs)
{
  int theTok;
  int id;
  char *loc;
  char *firstAddr;

  /* Weird Unix junk to ID the memory */
  theTok = ftok(path, 1);
  if (theTok == -1) {
    fprintf(stderr, "ftok(\"%s\", 1) failure\n");
    showErrno(errno);
    return 1;
    }

  /* estimate the overhead */
  /* overhead is 20 bytes per KB of cache size and 20KB per user */
  segSize += (segSize * 20) / 1000 + (20000 * numProcs);

  /* Create the memory */
  id = shmget(theTok, SHMGETCAST(segSize),
      SHM_PERMISSIONS | IPC_CREAT | IPC_EXCL);
  if (id == -1) {
    fprintf(stderr, "shmget() failure\n");
    showErrno(errno);
    return 1;
    }

  /* How odd.  Now we must attach to it, even though we created it. */
#if defined(FLG_SEQUENT_UNIX)
  firstAddr = (char*) 0x3000000;  /* to get larger break (heap limit) */
#else
  firstAddr = (char*) 0;
#endif
  loc = shmat(id, firstAddr, 0);
  if (loc == (char *)-1) {
    fprintf(stderr, "shmat() failure\n");
    showErrno(errno);
    /* _try_ to remove it */
    if (shmctl(id, IPC_RMID, NULL) < 0) {
      fprintf(stderr,
	  "painted into a corner!  shmctl(,IPC_RMID,) failed\n");
	  showErrno(errno);
      }
    return 1;
    }

  /* Undo everything now.  Try to detach. */
  if (shmdt(loc) < 0) {
    fprintf(stderr, "shmdt() failure\n");
    showErrno(errno);
    return 1;
    }

  /* Try to delete */
  if (shmctl(id, IPC_RMID, NULL) < 0) {
    fprintf(stderr, "shmctl(,IPC_RMID,) failed\n");
    showErrno(errno);
    return 1;
    }
  return 0;
}

/* the basic permissions for semaphores are not defined on all systems */
#if !defined(SEM_R)
#define SEM_R 0400  /* user read permission */
#endif
#if !defined(SEM_A)
#define SEM_A 0200  /* user alter permission */
#endif

/* the group permissions are not not defined at all in sem.h */
#define SEMG_R 0020 /* group read permission */
#define SEMG_A 0040 /* group alter permission */

/* define permissions used for our system */
#define SEM_PERMISSIONS  (SEM_R | SEM_A | SEMG_R | SEMG_A)


static int testSem(const char *path, int numSems)
{
  int theTok;
  int id;
  union semun semUnion;
  struct sembuf cmd;

  /* Weird Unix junk to ID the semaphore */
  theTok = ftok(path, 1);
  if (theTok == -1) {
    fprintf(stderr, "ftok(\"%s\", 1) failure\n");
    showErrno(errno);
    return 1;
    }

  /* Make the semaphore array */
  id = semget(theTok, numSems, SEM_PERMISSIONS | IPC_CREAT | IPC_EXCL);
  if (id == -1) {
    fprintf(stderr, "semget() failure path = %s, numSems = %d\n",
	path, numSems);
    showErrno(errno);
    return 1;
    }

  /* Fondle a semaphore, for kicks */
  cmd.sem_num = 1;
  cmd.sem_op = 1; /* clear */
  cmd.sem_flg = 0;
  if (semop(id, &cmd, 1) < 0) {
    fprintf(stderr, "semop() failer\n");
    showErrno(errno);
    if (semctl(id, 0, IPC_RMID, semUnion) < 0) {
      fprintf(stderr,
	  "painted in a corner!  semctl(, IPC_RMID) failed\n");
	  showErrno(errno);
      }
    return 1;
    }

  /* Destroy the array */
  memset((char *)&semUnion, 0, sizeof semUnion);
  if (semctl(id, 0, IPC_RMID, semUnion) < 0) {
    fprintf(stderr, "semctl(, IPC_RMID) failed\n");
    showErrno(errno);
    return 1;
    }

  /* success */
  return 0;
}


#if defined(MSDOS) || defined(macintosh)
int main()
#else
int main(int argc, char *argv[])
#endif
{
  int segSize;
  int semSize;
  int result = 0;
  
  if (argc != 4) {
    fprintf(stderr, "Usage: shmem <memfile> <kbytes> <numUsers>\n");
    fprintf(stderr, "    memfile is any file (used for getting an id)\n");
    fprintf(stderr, "    kbytes is the shared memory size\n");
    fprintf(stderr, "    numUsers is the number of GemStone users\n");

    fprintf(stderr, "Tests validity of shared memory and semaphores.\n");
    fprintf(stderr, "To test default configuration: 'shmem shmem 10000 45'\n");
    return 1;
    }

  if (access(argv[1], R_OK | W_OK) != 0) {
    fprintf(stderr, "access() error on %s\n", argv[1]);
    showErrno(errno);
    return 1;
    }

  segSize = atoi(argv[2]) * 1024;
  if (segSize <= 0) {
    fprintf(stderr, "illegal memsize\n");
    return 1;
    }

  semSize = atoi(argv[3]) + 1;
  if (semSize <= 0) {
    fprintf(stderr, "illegal semsize\n");
    return 1;
    }

  if (testMem(argv[1], segSize, semSize) != 0) {
    fprintf(stderr, "Unable to use shared memory, size %ld Kbytes\n",
	segSize / 1024);
    result = 1;
    }
  if (testSem(argv[1], semSize) != 0) {
    fprintf(stderr, "Unable to use semaphores, array size %d\n", semSize);
    result = 1;
    }

  return result;
}
